@@ -9,7 +9,7 @@ module Yesod.Core.Internal.TH where
 import Prelude hiding (exp)
 import Yesod.Core.Handler
 
-import Language.Haskell.TH hiding (cxt)
+import Language.Haskell.TH hiding (cxt, instanceD)
 import Language.Haskell.TH.Syntax
 
 import qualified Network.Wai as W
@@ -86,8 +86,13 @@ mkYesodGeneral namestr args isSub f resS = do
                    case info of
                      TyConI dec ->
                        case dec of
+#if MIN_VERSION_template_haskell(2,11,0)
+                         DataD _ _ vs _ _ _ -> length vs
+                         NewtypeD _ _ vs _ _ _ -> length vs
+#else
                          DataD _ _ vs _ _ -> length vs
                          NewtypeD _ _ vs _ _ -> length vs
+#endif
                          _ -> 0
                      _ -> 0
                _ -> return 0
@@ -101,7 +106,7 @@ mkYesodGeneral namestr args isSub f resS = do
                    case arg of
                      Left  t  -> ( ConT (mkName t):xs, n:ns, cs )
                      Right ts -> ( VarT n         :xs,   ns
-                                 , fmap (\t -> 
+                                 , fmap (\t ->
 #if MIN_VERSION_template_haskell(2,10,0)
                                                AppT (ConT $ mkName t) (VarT n)
 #else
@@ -164,7 +169,7 @@ mkDispatchInstance :: Type                      -- ^ The master site type
 mkDispatchInstance master cxt f res = do
     clause' <- mkDispatchClause (mkMDS f [|yesodRunner|]) res
     let thisDispatch = FunD 'yesodDispatch [clause']
-    return [InstanceD cxt yDispatch [thisDispatch]]
+    return [instanceD cxt yDispatch [thisDispatch]]
   where
     yDispatch = ConT ''YesodDispatch `AppT` master
 
@@ -181,3 +186,10 @@ mkYesodSubDispatch res = do
                     [innerFun]
                 ]
     return $ LetE [fun] (VarE helper)
+
+instanceD :: Cxt -> Type -> [Dec] -> Dec
+#if MIN_VERSION_template_haskell(2,11,0)
+instanceD = InstanceD Nothing
+#else
+instanceD = InstanceD
+#endif
